"""
Module for different solvers for eigenproblems

Current options: 

- scipy 
- pytorch (optional dependency, must be installed)

"""

from typing import Callable

import numpy as np
from scipy import sparse as sp

# Dict to lookup the solvers
__SOLVERS = {}

from .scipy_eigen import scipy_get_eigen

__SOLVERS["scipy"] = scipy_get_eigen

# Import guard the pytorch backend since it is optional
try:
    from .pytorch_eigen import torch_get_eigen

    __SOLVERS["pytorch"] = torch_get_eigen
    __SOLVERS["torch"] = torch_get_eigen
except (ImportError, RuntimeError):
    # Maybe display a warning?
    pass

# Function signature of the solver
Eigensolver = Callable[[sp.spmatrix, int, tuple[int]], tuple[np.ndarray, np.ndarray]]


def get_eigensolver(solver: str) -> Eigensolver | None:
    return __SOLVERS.get(solver)



# NOTE: this example uses skimage, available with:
# `pip install scikit-image`
from skimage import measure

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from qm_sim.hamiltonian import Hamiltonian
from qm_sim.nature_constants import e_0, 𝜀_0, m_e, m_p

N = (60, 60, 60)            # Discretisation point count
L = (1e-9, 1e-9, 1e-9)      # System size in meters
m = m_e * m_p / (m_e + m_p) # Mass of the particle, here the reduced mass of the system

nx, ny = 2, 2               # Used for the plot
n = nx * ny                 # Number of orbitals to find

# Initialize the hamiltonian.
H = Hamiltonian(N, L, m)

# Define the potential: -e^2 / (4*pi*𝜀_0*r)
x, y, z = np.meshgrid(*(np.linspace(-L[i]/2, L[i]/2, N[i]) for i in range(3)))
r = (x**2 + y**2 + z**2)**0.5
H.V = -e_0**2 / (4 * np.pi * 𝜀_0 * r)

# Find a couple eigenvalues
energies, orbitals = H.eigen(n)
orbitals = np.abs(orbitals**2)

# Plot a iso-surface of each orbital
plt.figure()
plots = []
for i in range(n):
    iso_val = np.mean(orbitals[i])
    vertices, faces, _, _ = measure.marching_cubes(orbitals[i], iso_val, spacing=L)
    ax = plt.subplot(ny, nx, i+1, projection='3d')
    plots.append(ax.plot_trisurf(vertices[:, 0], vertices[:,1], faces, vertices[:, 2],
                    cmap='Spectral', lw=1))
    plt.title(f"E = {energies[i] :.2e}")
plt.tight_layout()
plt.show()



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation
from scipy.special import expit

from qm_sim.hamiltonian import Hamiltonian
from qm_sim.nature_constants import e_0, m_e

s = 100
N = (s, s)              # Discretisation point count
L = (8e-9, 8e-9)        # System size in meters
m = m_e                 # Mass of the particle, here chosen as electron mass
n = 4                   # The amount of eigenstates to find
steps = 100


H = Hamiltonian(N, L, m)

# Set potential. Here, a square well is used, with dV = 0.15 eV
x,y = np.linspace(-L[0]/2,L[0]/2, N[0]), np.linspace(-L[0]/2,L[0]/2, N[1])
X, Y = np.meshgrid(x,y)

# Smoothed rectangular time-dependent potential, assymetric to avoid degeneracy 
def V(t):
    a = 0.15*e_0*expit(5*10**9*(X-2e-9 + 1e-11*t)) + 0.15*e_0*expit(-5*10**9*(X+2e-9-1e-11*t))
    b = 0.15*e_0*expit(5*10**9*(Y-2.5e-9 - 1e-11*t)) + 0.15*e_0*expit(-5*10**9*(Y+2.5e-9+1e-11*t))
    return np.where(a+b>0.15*e_0,0.15*e_0,a+b)

H.V = V

# Solve the system
energies, states = H.eigen(n)
# Choose energy to follow
E = energies[1]

# Evolve adiabatically
Energies, psi = H.adiabatic_evolution(E_n=E, t0=0, dt=1, steps=steps)

# Create animation
fig, (ax, ax2) = plt.subplots(1,2)
im = ax.matshow(psi[:,:,0]**2)
im2 = ax2.matshow(V(0).T)
ims = [im,im2]

def init():
    im.set_data(psi[:,:,0]**2)
    im2.set_data(V(0).T)
    return im

def update(j):
    ims[0].set_data(psi[:,:,j]**2)
    ims[1].set_data(V(j).T)
    return ims

anim = FuncAnimation(fig, func=update, init_func=init, frames=steps, interval=50, blit=False, repeat = True, repeat_delay = 1000)
plt.show()



"""
PyTorch backend for finding eigenstates
"""

import numpy as np
import torch
from scipy import sparse as sp

PYTORCH_DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def torch_get_eigen(mat: sp.dia_matrix, n: int, shape: tuple[int], **kwargs):
    """Calculate :code:`n` eigenvalues of :code:`mat`. Reshape output to :code:`shape`

    :param mat: Matrix to calculate eigenvectors and -values for
    :type mat: sp.dia_matrix
    :param n: Amount of eigenvectors and -values to calculate
    :type n: int
    :param shape: Output shape for eigenvectors
    :type shape: tuple[int]
    :return: eigenvalues, eigenvectors
    :rtype: tuple[np.ndarray(shape = (:code:`n`)), np.ndarray(shape = (:code:`n`, :code:`shape`)]
    """
    H = spmatrix_to_tensor(mat)
    v, w = torch.lobpcg(H, k=n, largest=False)
    w = w.cpu().numpy()
    w = np.array([w[:, i].reshape(shape, order="F") for i in range(n)])
    return v, w


def spmatrix_to_tensor(mat: sp.spmatrix) -> torch.Tensor:
    """
    Convert a sparse diagonal scipy matrix to a sparse pytorch tensor.
    """

    coo = mat.tocoo()
    values = coo.data
    indices = np.vstack((coo.row, coo.col))

    i = torch.from_numpy(indices)
    v = torch.from_numpy(values)
    shape = coo.shape

    return torch.sparse_coo_tensor(i, v, shape, device=PYTORCH_DEVICE)



import numpy as np
from scipy.integrate import solve_ivp

from .base import TemporalSolver


class ScipySolver(TemporalSolver):
    """Base class for scipy's :code:`solve_ivp`-based solvers"""

    #: Name of the :code:`solve_ivp` method
    method: str

    _skip_registration = True

    def iterate(
        self,
        v_0: np.ndarray,
        t0: float,
        t_final: float,
        dt: float,
        dt_storage: float = None,
        verbose: bool = True,
    ) -> tuple[np.ndarray, np.ndarray]:
        pbar = self.tqdm(t0, t_final, verbose)

        # Precalculate the coefficient for negligible speedup

        def ode_fun(t, y):
            pbar.progress(t)
            return self.H(t) @ y

        sol = solve_ivp(
            ode_fun,
            [t0, t_final],
            v_0.flatten(),
            t_eval=np.arange(t0, t_final, dt_storage),
            first_step=dt,
            method=self.method,
        )
        t = sol.t
        psi = [sol.y[:, i].reshape(*self.output_shape) for i in range(len(t))]

        return t, np.array(psi)

    def __init_subclass__(cls):
        cls.name = "scipy-" + cls.name
        cls._skip_registration = False
        return super().__init_subclass__()


class RungeKutta45(ScipySolver):
    order = 5
    explicit = True
    stable = True
    name = "Runge-Kutta 5(4)"
    method = "RK45"


class RungeKutta23(ScipySolver):
    order = 3
    explicit = True
    stable = True
    name = "Runge-Kutta 3(2)"
    method = "RK23"


class DOP853(ScipySolver):
    order = 8
    explicit = True
    stable = True
    name = "DOP853"
    method = "DOP853"


# Solver does not support complex numbers
# class Radau(ScipySolver):
#     order = 5
#     explicit = False
#     stable = True
#     name = "radau"
#     method = "Radau"


class BDF(ScipySolver):
    order = None
    explicit = False
    stable = True
    name = "backwards-differentiation"
    method = "BDF"


# Solver does not support complex numbers
# class LSODA(ScipySolver):
#     order = None
#     explicit = None
#     stable = None
#     name = "fortran LSODA"
#     method = "LSODA"



"""
Different Hamiltonian classes.

Available options:

- Hamiltonian (proxy of SpatialHamiltonian)
- SpatialHamiltonian

"""

# The spatial Hamiltonian is the default
from .spatial_hamiltonian import SpatialHamiltonian
from .spatial_hamiltonian import SpatialHamiltonian as Hamiltonian



import numpy as np

from qm_sim.hamiltonian import Hamiltonian
from qm_sim.nature_constants import e_0, m_e
from scipy.signal import convolve2d

N = (100, 100)          # Discretisation point count
L = (15e-9, 15e-9)      # System size in meters
m = m_e                 # Mass of the particle, here chosen as electron mass
t_end = 100e-15         # Simulation time
dt = 1e-16              # Simulation data storage stepsize

# Initialize the hamiltonian.
H = Hamiltonian(N, L, m, temporal_scheme="scipy-Runge-Kutta 3(2)")

# Set the potential. Use a cool elipse cross that rotates over time
a = 2e-9
b = 5e-9
x = np.linspace(-L[0]/2, L[0]/2, N[0])
y = np.linspace(-L[1]/2, L[1]/2, N[1])
X, Y = np.meshgrid(x, y)
def V(theta: float) -> np.ndarray:
    ct, st = np.cos(theta), np.sin(theta)

    V = np.ones(N) * e_0
    V[((X*ct + Y*st) / a)**2 + ((X*st - Y*ct) / b)**2 <= 1] = 0
    V[((X*ct + Y*st) / b)**2 + ((X*st - Y*ct) / a)**2 <= 1] = 0

    # Smooth out transition 
    n = 5
    kernel = np.ones((n,n)) / n**2
    V = convolve2d(V, kernel, mode="same", boundary="wrap")
    return V
    
H.V = lambda t: V(1e14*t)

# Set initial condition to a
# superposition of third and fourth eigenstate
_, psi = H.eigen(4)
psi_0 = 1/2**0.5 * (psi[2] + psi[3])

# Plot
H.plot_potential()
H.plot_eigen(4)
H.plot_temporal(t_end, dt, psi_0)



"""

Calculate eigenenergies and eigenstates of a 1D system with zero potential

"""
from qm_sim.hamiltonian import Hamiltonian
from qm_sim.nature_constants import m_e

N = 100         # Discretisation point count
L = 1e-9        # System size in meters
m = m_e         # Mass of the particle, here chosen as electron mass
n = 4           # The amount of eigenstates to find

# Set hamiltonian
H = Hamiltonian(N, L, m)

# Solve and plot the system
H.plot_eigen(n)



from qm_sim.cpp import eigen as s
from matplotlib import pyplot as plt
import numpy as np
from qm_sim.hamiltonian import Hamiltonian
from qm_sim.nature_constants import m_e, e_0
import time
import functools

from .scipy_backend import TestHam, eigs, eigsh

# Decorator to print the time a function took
# Optionally give an amount of calls to the function
def timer(n: int = 1):
    def decorator(func):
        @functools.wraps(func)
        def _timer(*args, **kwargs):
            start = time.perf_counter()
            for _ in range(n):
                out = func(*args, **kwargs)
            print(f"{func.__name__}\n\ttook {time.perf_counter() - start :#.5g}s")
            return out
        return _timer
    return decorator

def test_eigensolvers():

    N = [1000,]
    k = 4
    L = [10.0,]
    stencil = [-2.0, 1.0]
    n_iter = 10

    V = [0.5 for _ in range(N[0])]
    for i in range(N[0]//4, 3*N[0]//4):
        V[i] = 0.0
    
    H = s.Hamiltonian(N, L, stencil, 1.0)
    H.set_potential(V)

    H2 = TestHam(N, L, stencil, 1.0)
    H2.set_potential(V)
    H2_matop = H2.matop()

    # NOTE: the Hamiltonian class does NOT use reduced units (yet)
    H3 = Hamiltonian(N, [1e-9*i for i in L], m_e)
    H3.V = np.array(V) * e_0

    H4 = Hamiltonian(N, [1e-9*i for i in L], m_e, eigensolver="pytorch")
    H4.V = np.array(V) * e_0

    @timer(n_iter)
    def spectra():
        return s.eigen(N, L, stencil, k, 1.0)

    @timer(n_iter)
    def scipy_cpp_op():
        return eigsh(H.as_operator, N[0], a1.dtype, which="SA", k=k)

    @timer(n_iter)
    def scipy_numba_op():
        return eigsh(H2_matop, N[0], a1.dtype, which="SA", k=k)

    @timer(n_iter)
    def scipy_numba_op_nonsymmetric():
        return eigs(H2_matop, N[0], a1.dtype, which="SR", k=k)
    
    @timer(n_iter)
    def scipy_matrix_op():
        return eigsh(H3.mat._mul_vector, N[0], a1.dtype, which="SA", k=k)

    @timer(n_iter)
    def scipy_matrix_op_nonsymmetric():
        return eigs(H3.mat._mul_vector, N[0], a1.dtype, which="SR", k=k)

    @timer(n_iter)
    def qm_sim():
        return H3.eigen(k)
    
    @timer(n_iter)
    def qm_sim_gpu():
        return H4.eigen(k)

    a1 = spectra()[:, ::-1]
    w, a2 = scipy_cpp_op()
    w, a3 = scipy_numba_op()
    _, a5 = scipy_matrix_op()
    scipy_matrix_op_nonsymmetric()
    scipy_numba_op_nonsymmetric()
    _, a4 = qm_sim()
    _, a6 = qm_sim_gpu()


    plt.figure()
    plt.subplot(1, 2, 1)
    plt.plot(a1[:, 1])
    plt.plot(a2[:, 1])
    plt.plot(a3[:, 1])
    plt.plot(a5[:, 1])
    # plt.plot(a4[1, :]) # normalised => not on the same scale as the rest
    plt.legend(["spectra", "scipy cpp", "scipy numba", "scipy python", "qm_sim",])
    plt.subplot(1, 2, 2)
    plt.plot(a1[:, 3])
    plt.plot(a2[:, 3])
    plt.plot(a3[:, 3])
    plt.plot(a5[:, 3])
    # plt.plot(a4[3, :]) # normalised => not on the same scale as the rest
    plt.legend(["spectra", "scipy cpp", "scipy numba", "scipy python", "qm_sim",])
    plt.tight_layout()
    plt.show()

    plt.figure()
    plt.subplot(1, 2, 1)
    plt.plot(a2)
    plt.legend(["cpp-scipy"])
    plt.subplot(1, 2, 2)
    plt.plot(a3)
    plt.legend(["numba-scipy"])
    plt.show()

    plt.figure()
    plt.plot(w[2]*a3[:, 2])
    plt.plot(H2.matop()(a3[:, 2]))
    plt.title("Totally an eigenvector guys cmon")
    plt.legend(["$\lambda v$", "$Av$"])
    plt.show()

def test_implemented_eigensolvers():

    from qm_sim.hamiltonian import Hamiltonian
    N = (200, 200)
    k = 4
    L = (10e-9, 10e-9)
    n_iter = 1

    V = e_0 * np.ones(N)

    H1 = Hamiltonian(N, L, m_e, eigensolver="scipy")
    H1.V = V

    H2 = Hamiltonian(N, L, m_e, eigensolver="pytorch")
    H2.V = V

    @timer(n_iter)
    def scipy_backend():
        return H1.eigen(k)
    
    @timer(n_iter)
    def pytorch_backend():
        return H2.eigen(k)
    
    scipy_backend()
    pytorch_backend()



"""

Calculate eigenenergies and eigenstates of a 1D system with quadratic potential

"""
from qm_sim.hamiltonian import Hamiltonian
from qm_sim.nature_constants import m_e

import numpy as np

N = 1000        # Discretisation point count
L = 1e-9        # System size in meters
m = m_e         # Mass of the particle, here chosen as electron mass
n = 4           # The amount of eigenstates to find

# Set hamiltonian
H = Hamiltonian(N, L, m)

# Set potential
x = np.linspace(-L/2, L/2, N)

def V(t):
    k = 200
    return 1/2*(k+t)*x**2
H.V = V

# Plot
H.plot_potential()
H.plot_eigen(n)



"""

Classes to solve :math:`y' = f(y)`

"""


from .base import TemporalSolver, get_temporal_solver
from .crank_nicolson import CrankNicolson
from .leapfrog import Leapfrog
from .scipy_solvers import BDF, DOP853, RungeKutta23, RungeKutta45



from typing import Callable

import numpy as np
from scipy.linalg import solve_banded
from scipy.sparse import dia_matrix

from .base import TemporalSolver


def I_plus_aH(a: complex, H: dia_matrix) -> dia_matrix:
    """Return I + aH, where :code:`I`is the unit matrix and :code:`a` is constant"""
    out = H.copy().astype(np.complex128)
    out.data *= a
    zero_ind = list(out.offsets).index(0)
    out.data[zero_ind] += 1
    return out


class CrankNicolson(TemporalSolver):
    order = 2
    explicit = False
    stable = True
    name = "crank-nicolson"

    def __init__(
        self, H: Callable[[float], np.ndarray], output_shape: tuple[int] = None
    ):
        TemporalSolver.__init__(self, H, output_shape)
        if len(self.output_shape) != 1:
            raise ValueError("Crank-Nicolson solver only supports 1D systems")

    def iterate(
        self,
        v_0: np.ndarray,
        t0: float,
        t_final: float,
        dt: float,
        dt_storage: float = None,
        verbose: bool = True,
    ) -> tuple[np.ndarray, np.ndarray]:
        if dt_storage is None:
            dt_storage = dt

        H = self.H
        prefactor = dt / 2

        tn = t0
        psi_n = v_0

        t = [tn]
        psi = [psi_n]

        with self.tqdm(t0, t_final, verbose) as pbar:
            while tn < t_final:
                # psi^n+1 = psi^n + dt/2*(F^n+1 + F^n)
                # F^n = 1/ihbar * H^n @ psi^n
                # => psi^n+1 = psi^n + dt/2ihbar*(H^n+1 @ psi^n+1 + H^n @ psi^n)
                # (I - dt/2ihbar*(H^n+1) @ psi^n+1 = (I + dt/2ihbar*H^n) @ psi^n

                lhs = I_plus_aH(-prefactor, H(tn + dt))
                rhs = I_plus_aH(prefactor, H(tn)) @ psi_n

                # Reformulate the solve_banded function in terms of the dia_matrix class
                psi_n = solve_banded(
                    (-lhs.offsets[0], lhs.offsets[-1]),
                    lhs.data[::-1, :],
                    rhs,
                )

                pbar.update(dt)
                tn += dt

                # store data every :code:`dt_storage`seconds
                if tn // dt_storage > len(psi):
                    psi.append(psi_n)
                    t.append(tn)
        return np.array(t), np.array(psi)



from abc import ABC, abstractmethod
from typing import Callable

import numpy as np
from tqdm import tqdm

from ..nature_constants import h_bar

# Dict to store subclasses of TemporalSolver
_SCHEMES = {}


class TemporalSolver(ABC):
    """Class to solve :math:`y' = H(y)`"""

    #: Name of the solver
    name: str

    #: Integration order
    order: int

    #: Is the method explicit or implicit?
    explicit: bool

    #: Is the method stable?
    #: If only conditionally stable, this will be true
    #: and :code:`dt` will be forced into its stable range
    stable: bool

    _skip_registration: bool = False

    def __init__(
        self, H: Callable[[float], np.ndarray], output_shape: tuple[int] = None
    ):
        """Initialize a temporal solver

        :param H: Function of time, representing the temporal derivative at that time
        :type H: Callable[[float], np.ndarray]
        :param output_shape: Expected shape of the solution, defaults to None
        :type output_shape: tuple[int], optional
        """

        self.H = H

        if output_shape is None:
            output_shape = self.H(0).shape
        self.output_shape = output_shape

    def tqdm(self, t_start: float, t_end: float, enable: bool) -> tqdm:
        pbar = tqdm(
            desc=self.name + " solver", total=t_end - t_start, disable=not enable
        )
        pbar.bar_format = "{l_bar}{bar}| {n:#.02g}/{total:#.02g}"

        # Add func to update with t, not dt
        pbar.progress = lambda t: pbar.update(t - pbar.n)

        return pbar

    @abstractmethod
    def iterate(
        self,
        v_0: np.ndarray,
        t0: float,
        t_final: float,
        dt: float,
        dt_storage: float = None,
        verbose: bool = True,
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Iterate the time propagation scheme.
        Store the current state every :code:`dt_storage`

        Args:
            t_final (float):
                End time for calculations
            dt_storage (float, optional):
                Data storage period.
                If None, store each calculation :code:`dt`
                Defaults to None.

        Returns:
            np.ndarray:
                Time values, shape (:code:`n`,) for :code:`n` storage times
            np.ndarray:
                State at times stored in the other output. shape (:code:`n`, :code:`H`.shape)
        """
        pass

    def __init_subclass__(cls):
        """Register subclasses of :class:`TemporalSolver`"""
        if cls._skip_registration:
            return
        # Register new subclasses of TemporalSolver
        if _SCHEMES.get(cls.name) is None:
            _SCHEMES[cls.name] = cls
        else:
            raise ValueError("Cannot have two schemes with the same name")


def get_temporal_solver(scheme: str) -> type[TemporalSolver]:
    """Get a solver from its name, if it exists

    :param scheme: Name of the solver
    :type scheme: str
    :raises ValueError: If the scheme does not exist
    :return: A temporal solver class, NOT an instance
    :rtype: type[TemporalSolver]
    """
    if scheme in _SCHEMES.keys():
        return _SCHEMES[scheme]
    raise ValueError(
        f"Scheme {scheme} not found. Options are:\n" + "\n".join(_SCHEMES.keys())
    )



import numpy as np

from qm_sim.hamiltonian import Hamiltonian
from qm_sim.nature_constants import e_0, m_e


N = 200                 # Discretisation point count
L = 2e-9                # System size in meters
m = m_e                 # Mass of the particle, here chosen as electron mass
t_end = 10e-15          # Simulation time
dt = 1e-17              # Simulation data storage stepsize

# Initialize the hamiltonian
H = Hamiltonian(N, L, m, temporal_scheme="leapfrog")

# Set the potential to a quadratic potential oscilating from side to side
z = np.linspace(-L/2, L/2, N)
H.V = lambda t: 6*z**2 + 3*z*np.abs(z)*np.sin(4e15*t)

# Plot
H.plot_temporal(t_end, dt)



"""

Nature constants, in SI units.

Source: https://physics.nist.gov/cuu/Constants/index.html 

"""

# Exact values
e_0         =       1.602176634e-19         # [C]   Elementary charge
h           =       6.62607015e-34          # [JS]  Planck constant, exact
k_b         =       1.380649e-23            # [J/K] Boltzmann constant, exact

# Measured values
m_e         =       9.1093837015e-31        # [kg]  Electron mass
m_p         =       1.67262192369e-27       # [kg]  Proton mass
h_bar       =       1.054571817e-34         # [Js]  Reduced Planck constant  
𝜀_0         =       8.8541878128e-12        # [F/m] Vacuum permittivity



"""

Discretized :math:`\\nabla` and :math:`\\nabla^2`

"""

_SCHEMES = {
    "three-point": 2,
    "five-point": 4,
    "seven-point": 6,
    "nine-point": 8,
}


def get_scheme_order(scheme: str) -> int | None:
    return _SCHEMES.get(scheme)



import numpy as np

from .base import TemporalSolver


class Leapfrog(TemporalSolver):
    order = 2
    explicit = True
    stable = True  # conditionally stable, dt is chosen accordingly
    name = "leapfrog"

    def iterate(
        self,
        v_0: np.ndarray,
        t0: float,
        t_final: float,
        dt: float,
        dt_storage: float = None,
        verbose: bool = True,
    ) -> tuple[np.ndarray, np.ndarray]:
        # Override initial condition to preserve 2nd order accuracy
        # psi_0 = psi_half - c*H_half*psi_half
        # psi_1 = psi_half + c*H_half*psi_half
        # c = dt / (2i*hbar)

        if dt_storage is None:
            dt_storage = dt

        H = self.H

        psi_half = v_0.flatten()
        psi_0 = psi_half - dt / 2 * (H(t0 + dt / 2) @ psi_half)
        psi_1 = psi_half + dt / 2 * (H(t0 + dt / 2) @ psi_half)

        tn = t0 + dt / 2

        psi = [psi_1.reshape(self.output_shape)]
        t = [tn]
        with self.tqdm(t0, t_final, verbose) as pbar:
            while tn < t_final:
                # psi^n+1 = psi^n-1 + 2*dt*F^n
                # F^n = 1/ihbar * H^n @ psi^n
                # H^n = H0 + V^n

                psi_2 = 2 * dt * (H(tn) @ psi_1) + psi_0
                psi_0, psi_1 = psi_1, psi_2

                pbar.update(dt)
                tn += dt

                # store data every :code:`dt_storage`seconds
                if tn // dt_storage > len(psi):
                    psi.append(psi_0.reshape(self.output_shape))
                    t.append(tn)
        return np.array(t), np.array(psi)



import numba
import numpy as np

from typing import Callable

from scipy._lib._threadsafety import ReentrancyLock
from scipy.sparse.linalg._eigen.arpack.arpack import _SymmetricArpackParams, _UnsymmetricArpackParams


def eigsh(A_OP, n, dtype, k=6, sigma=None, which='LM', v0=None,
          ncv=None, maxiter=None, tol=0, return_eigenvectors=True,
          ):
    mode = 1
    M_matvec = None
    Minv_matvec = None
    params = _SymmetricArpackParams(n, k, dtype.char, A_OP, mode,
                                    M_matvec, Minv_matvec, sigma,
                                    ncv, v0, maxiter, which, tol)

    with ReentrancyLock("Nested calls to eigs/eighs not allowed: "
        "ARPACK is not re-entrant"):
        while not params.converged:
            params.iterate()

        return params.extract(return_eigenvectors)

def eigs(A_OP, n, dtype, k=6, sigma=None, which='LM', v0=None,
          ncv=None, maxiter=None, tol=0, return_eigenvectors=True,
          ):
    mode = 1
    M_matvec = None
    Minv_matvec = None
    params = _UnsymmetricArpackParams(n, k, dtype.char, A_OP, mode,
                                    M_matvec, Minv_matvec, sigma,
                                    ncv, v0, maxiter, which, tol)

    with ReentrancyLock("Nested calls to eigs/eighs not allowed: "
        "ARPACK is not re-entrant"):
        while not params.converged:
            params.iterate()

        return params.extract(return_eigenvectors)

class TestHam:
    def __init__(self, N, L, stencil, m = 1.0):
        self.N = np.array(N)
        self.L = np.array(L)
        self.stencil = np.array(stencil)
        self.m = m
        self.ndim = len(N)
        self.V = np.zeros(self.n)
        self.h0 = np.array([-0.0380998212 / (m * (l/n)**2) for l, n in zip(L, N)])
    
    def set_potential(self, V: list | np.ndarray):
        if isinstance(V, list):
            self.V = np.array(V)
        else:
            self.V = V.copy()

    @property
    def n(self) -> int:
        res = 1
        for n in self.N:
            res *= n
        return res
    
    def matop(self) -> Callable[[np.ndarray], np.ndarray]:
        V = self.V
        ndim = self.ndim
        h0 = self.h0
        n = self.n
        stencil = self.stencil
        N = self.N

        @numba.njit(parallel=True)
        def scipy_numba_matop(x_in: np.ndarray) -> np.ndarray:

            y_out = V.copy()

            offset = 1
            
            for dim in range(ndim):

                #  1 / dx^2
                h0_dim = h0[dim]

                #  y_i = x_i (*) stencil
                #  stencil usually has more terms, so it can be e.g.
                #  y_i = stencil[1]*x_i-1 + stencil[0]*x_i + stencil[1]*x_i+1
                for i in numba.prange(n):
                    stencil_sum = x_in[i] * stencil[0]
                    for j, el in enumerate(stencil):
                        # enumerate(stencil, start=1) fails with jit
                        if j == 0:
                            continue
                        #  account for current dimension
                        j *= offset
                        #  range-check the x-vector
                        if ((i + j) < n):
                            stencil_sum += x_in[i + j] * el
                        
                        if ((i - j) >= 0):
                            stencil_sum += x_in[i - j] * el
                        
                    y_out[i] += h0_dim * stencil_sum

                offset *= N[dim]

            return y_out
        
        # compile
        scipy_numba_matop(np.zeros(self.N).flatten())
        
        return scipy_numba_matop

    



"""

Real-space discretized Hamiltonian class, with solving and plotting functionality

"""

from collections.abc import Iterable
from typing import Any, Callable

import numpy as np
from scipy.sparse import dia_matrix
from scipy.sparse.linalg import eigsh as adiabatic_eigsh
from tqdm import tqdm

from .. import plot
from ..eigensolvers import get_eigensolver
from ..nature_constants import h_bar
from ..spatial_derivative import get_scheme_order
from ..spatial_derivative.cartesian import laplacian, nabla
from ..temporal_solver import TemporalSolver, get_temporal_solver


class SpatialHamiltonian:
    def __init__(
        self,
        N: tuple[int] | int,
        L: tuple[float] | float,
        m: float | np.ndarray,
        spatial_scheme: str = "three-point",
        temporal_scheme: str = "leapfrog",
        eigensolver: str = "scipy",
        verbose: bool = True,
        boundary_condition: str = "zero",
    ):
        """Non-stationary Hamiltonian in real space.

        :param N: Discretization count along each axis
        :type N: tuple[int] | int
        :param L: System size along each axis
        :type L: tuple[float] | float
        :param m: Mass of the particle in the system.
            Can be constant (float) or vary in the simulation area (array).
            If an array is used, :code:`m.shape == N` must hold
        :type m: float | np.ndarray
        :param spatial_scheme: Finite difference scheme for spatial derivative.
            Options are:

            - three-point
            - five-point
            - seven-point
            - nine-point

            Defaults to "three-point"
        :type spatial_scheme: str, optional
        :param temporal_scheme: Finite difference scheme for temporal derivative.
            Options are:

            - crank-nicolson
            - leapfrog
            - scipy-Runge-Kutta 5(4)
            - scipy-Runge-Kutta 3(2)
            - scipy-DOP853
            - scipy-backwards-differentiation

            Defaults to "leapfrog"
        :type temporal_scheme: str, optional
        :param eigensolver: Choose which eigensolver backend to use.
            Options are:

            - scipy
            - torch (optional dependency, must be installed)

            Defaults to "scipy"
        :type eigensolver: str, optional
        :param verbose: Option to display calculation and iteration info during runtime.
            Defaults to True
        :type verbose: bool, optional
        :param boundary_condition: Which boundary condition to apply.
            Options are:

            - zero
            - periodic

            Defaults to "zero"
        :type boundary_condition: str, optional
        """
        # 1D inputs
        if isinstance(N, int):
            N = (N,)
        if isinstance(L, (float, int)):
            L = (L,)

        # Allow any iterable that can be converted to a tuple
        if isinstance(N, Iterable):
            N = tuple(N)
        if isinstance(L, Iterable):
            L = tuple(L)

        # Check type
        if not isinstance(N, tuple) or not all(isinstance(i, int) for i in N):
            raise ValueError(f"Param `N` must be int or tuple of ints, got {type(N)}")
        if not isinstance(L, tuple) or not all(isinstance(i, (float, int)) for i in L):
            raise ValueError(f"Param `L` must be float or tuple, got {type(L)}")

        if len(N) != len(L):
            raise ValueError("`N`and `L`must have same length")

        self.N = N
        self.L = L

        self.eigensolver = get_eigensolver(eigensolver)
        if self.eigensolver is None:
            raise ValueError(f"Eigensolver {eigensolver} not found")

        order = get_scheme_order(spatial_scheme)
        if order is None:
            raise ValueError("Requested finite difference is invalid")

        self._dim = len(N)
        self.delta = [Li / Ni for Li, Ni in zip(L, N)]

        # Handle non-isotropic effective mass
        if isinstance(m, np.ndarray) and np.all(m == m.flat[0]):
            m = m.flatten()[0]
        if isinstance(m, np.ndarray):
            print("Warning: Continuity is NOT satisfied (yet) with non-isotropic mass")
            if m.shape != self.N:
                raise ValueError(
                    f"Inconsistent shape of `m`: {m.shape}, should be {self.N}"
                )
            m_inv = 1 / m.flatten()

            _n = nabla(N, L, order=order, boundary_condition=boundary_condition)
            _n2 = laplacian(N, L, order=order, boundary_condition=boundary_condition)

            # nabla m_inv nabla + m_inv nabla^2
            _n.data *= _n @ m_inv  # First term
            _n2.data *= m_inv  # Second term
            self.mat = _n + _n2
        else:
            self.mat = laplacian(
                N, L, order=order, boundary_condition=boundary_condition
            )
            self.mat *= 1 / m

        self._centerline_index = list(self.mat.offsets).index(0)
        self._default_data = None

        # Multiply the laplacian with the remaining factors
        self.mat.data *= -(h_bar**2) / 2

        # static zero potential by default
        self._V = lambda t: np.zeros(shape=N)

        self.verbose = verbose

        self._temporal_solver = get_temporal_solver(temporal_scheme)

    @property
    def V(self) -> Callable[[float], np.ndarray]:
        """Potential as a function of time

        :return: input time, output potential. Output must have same shape as systme
        :rtype: Callable[[float], np.ndarray]
        """
        return self._V

    @V.setter
    def V(self, V: np.ndarray | Callable[[float], np.ndarray]):
        """Set a (potentially time dependent) potential for the QM-system's Hamiltonian

        :param V: The energetic value of the potential at a given point associated with
            given array indices,
            if callable, the call variable will represent a changable parameter
            (usually time) with a return type identical to the static case where V is an np.ndarray
        :type V: np.ndarray | Callable[[float], np.ndarray]
        """
        if not callable(V):
            # Avoid the lambda func referring to itself
            array_V = V
            V = lambda t: array_V

        if V(0).shape != self.shape:
            raise ValueError(f"Inconsistent shape. Shape must be {self.shape}")

        self._V = V

    def eigen(self, n: int, t: float = 0, **kwargs) -> tuple[np.ndarray, np.ndarray]:
        """Calculate the n smallest eigenenergies and the corresponding eigenstates of
        the hamiltonian

        :param n: Amount of eigenenergies/states to output
        :type n: int
        :param t: If the potential is time-dependent, solves the Time-independent
            Schrödinger eq. as if it was frozen at time t.
            Does nothing if potential is time-independent.
            Defaults to 0
        :type t: float, optional
        :return:
            - Eigenenergies
            - Normalised eigenstates

        :rtype: tuple[np.ndarray(shape = (n)), np.ndarray(shape = (n, N))]
        """

        # The adiabatic solver uses some features of the eigensolver
        # not exposed by the `self.eigensolver` function
        if kwargs.pop("is_adiabatic", False):
            E, psi = adiabatic_eigsh(self(t), k=n, **kwargs)
            psi = np.array([psi[:, i].reshape(self.N, order="F") for i in range(n)])
        else:
            E, psi = self.eigensolver(self(t), n, self.N)

        # calculate normalisation factor
        normalisation_factor = [psi[i, :] ** 2 for i in range(n)]
        for i, (L, N) in enumerate(zip(self.L, self.N)):
            dx = L / N
            for j in range(n):
                normalisation_factor[j] = np.trapz(normalisation_factor[j], dx=dx)
        # normalise
        for i in range(n):
            psi[i] /= normalisation_factor[i] ** 0.5
        return E, psi

    def adiabatic_evolution(
        self, E_n: float, t0: float, dt: float, steps: int
    ) -> tuple[np.ndarray, np.ndarray]:
        """Adiabatically evolve an eigenstate with a slowly varying time-dependent potential with
        energy (close to) :code:`E_n` using the Adiabatic approximation.
        https://en.wikipedia.org/wiki/Adiabatic_theorem

        Note: This is only valid given that the adiabatic theorem holds, ie. no degeneracy and a
        gap betweeneigenvalues. Current implementation assumes this holds and does not check if
        it does (yet?).
        There is no mathematical guarantee (yet?) that the iterative solver will "hug" the
        correct eigenvector at every step, but it should be good if V varies smoothly enough and
        :code:`dt` is small enough.


        :param E_n: The Eigenvalue for you want to adiabatically evolve
        :type E_n: float
        :param t0: Starting time for time-evolution
        :type t0: float
        :param dt: The (small) time parameter increment used to update the eigenstate temporally
        :type dt: float
        :param steps: Number of increments.
        :type steps: int
        :return: (E(t), Psi(t)), Time evolution of the eigenstate.
        :rtype: tuple[np.ndarray(shape = steps), np.ndarray(shape = (N, steps))]
        """
        Psi_t = np.empty(shape=(*self.N, steps + 1))
        En_t = np.empty(steps + 1)
        t = t0

        # initialize
        _, Psi_t[:, :, 0] = self.eigen(1, t, sigma=E_n, is_adiabatic=True)
        En_t[0] = E_n

        for i in tqdm(
            range(1, steps + 1), desc="Adiabatic evolution", disable=not self.verbose
        ):
            t += dt
            En_t[i], Psi_t[:, :, i] = self.eigen(
                1,
                t,
                # smartly condition eigsolver to "hug" the single eigenvalue solution;
                # eigenvector and eigenvalue should be far closer to the previous one
                # than any other if the adiabatic theorem is fulfilled
                sigma=En_t[i - 1],
                v0=-Psi_t[:, :, i - 1],
                is_adiabatic=True,
            )
        return En_t, Psi_t

    def temporal_evolution(
        self,
        t0: float,
        t_final: float,
        dt_storage: float = None,
        psi_0: np.ndarray = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        # Default: superposition of 1st and 2nd eigenstate
        if psi_0 is None:
            _, psi = self.eigen(2)
            psi_0 = 2**-0.5 * (psi[0] + psi[1])

        # Calculate a good dt from von Neumann analysis of the leapfrog scheme
        # NOTE: this assumes the temporal part is at most 4x the static part,
        #       and that the potential takes its maximum somwhere at t=t0
        V_max = np.max(self.V(t0) * 4)
        V_min = np.min(self.V(t0) * 4)

        E_max = max(
            abs(V_min),
            abs(V_max + 4 * np.max(self.mat.data)),
        )
        dt = 0.25 * h_bar / E_max

        # solve
        func = lambda t: 1 / (1j * h_bar) * self.__call__(t)
        solver = self._temporal_solver(func, self.shape)
        t, psi = solver.iterate(
            psi_0.astype(np.complex128), t0, t_final, dt, dt_storage, self.verbose
        )

        return t, psi

    temporal_evolution.__doc__ = TemporalSolver.iterate.__doc__

    def plot_eigen(self, n: int, t: float = 0):
        """Calculate and plot :code:`n` eigenstates at time :code:`t`

        :param n: Amount of eigenstates to plot
        :type n: int
        :param t: Time at which to find eigenstates, defaults to 0
        :type t: float, optional
        """
        E, psi = self.eigen(n, t)
        plot.eigen(E, psi)

    def plot_temporal(
        self, t_final: float, dt: float, psi_0: np.ndarray = None, t0: float = 0
    ):
        """Plot the temporal evolution of the eigenstates

        :param t_final: Simulation end time
        :type t_final: float
        :param dt: Simulation time between each timestep
        :type dt: float
        :param psi_0: Initial state, defaults to None
        :type psi_0: np.ndarray, optional
        :param t0: Simulation start time, defaults to 0
        :type t0: float, optional
        """
        t, psi = self.temporal_evolution(t0, t_final, dt, psi_0)
        plot.temporal(t, psi, self.V)

    def plot_potential(self, t: float = 0):
        """Plot the potential at a given time

        :param t: Time at which to plot, defaults to 0
        :type t: float, optional
        """
        plot.potential(self.V(t))

    def __add__(self, other: np.ndarray) -> dia_matrix:
        """Add a vector to the main diagonal, and return the matrix

        :param other: Array to add to the main diagonal
        :type other: np.ndarray
        :return: Matrix data, i.e. NOT a Hamiltonian object
        :rtype: dia_matrix
        """
        if self._default_data is None:
            self._default_data = self.mat.data.copy()

        self.mat.data = self._default_data.copy()
        self.mat.data[self._centerline_index, :] += other.flatten()
        return self.mat

    def __call__(self, t: float) -> dia_matrix:
        """Get the matrix at a given time

        :param t: Time at which to get the matrix at
        :type t: float
        :return: Matrix data, i.e. NOT a Hamiltonian object
        :rtype: dia_matrix
        """
        return self + self.V(t)

    @property
    def shape(self) -> tuple[int]:
        """System shape

        :return: Tuple with discretization count along each axis
        :rtype: tuple[int]
        """
        return self.N

    @property
    def N_total(self) -> int:
        """Total amount of discretization points

        :return: Discretization point count
        :rtype: int
        """
        i = 1
        for j in self.N:
            i *= j
        return i

    @property
    def ndim(self) -> int:
        """System dimensionallity

        :return: Amount of dimensions in the system
        :rtype: int
        """
        return len(self.N)

    def __matmul__(self, other: Any) -> Any:
        """Matrix multiplication with the underlying data

        :param other: matrix to multiply with
        :type other: Any
        :return: Matrix product
        :rtype: Any
        """
        return self.mat @ other

    def _fast_matmul_op(self, t: float = 0) -> Callable[[Any], np.ndarray]:
        """Evaluate the Hamiltonian at a given time,
        and return a matrix-vector multiplication function

        :param t: Time at which to evaluate the Hamiltonian, defaults to 0
        :type t: float, optional
        :return: Fast matrix-vector multiplication function
        :rtype: Callable[[Any], np.ndarray]
        """
        mat = self(t)
        return mat._mul_vector

    def asarray(self) -> np.ndarray:
        """Return the data matrix as a dense array

        :return: Dense array of the discretized Hamiltonian
        :rtype: np.ndarray
        """
        return self.mat.toarray()



"""

Calculate and plot eigenstates in a 2D potential

"""
from qm_sim.hamiltonian import Hamiltonian
from qm_sim.nature_constants import m_e, e_0

import numpy as np

N = (200, 200)          # Discretisation point count
L = (8e-9, 8e-9)        # System size in meters
m = m_e                 # Mass of the particle, here chosen as electron mass
n = 4                   # The amount of eigenstates to find

# Set hamiltonian
H = Hamiltonian(N, L, m)

# Set potential. Here, a square well is used, with dV = 0.15 eV
V = 0.15*e_0 * np.ones(N)
V[
    N[0] // 2 - N[0] // 5 : N[0] // 2 + N[0] // 5, 
    N[1] // 2 - N[1] // 5 : N[1] // 2 + N[1] // 5
] = 0
H.V = V

# Plot potential
H.plot_potential()

# Plot eigenstate
H.plot_eigen(n)




import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation
from scipy.signal import convolve

from qm_sim.hamiltonian import Hamiltonian
from qm_sim import nature_constants as const


def test_constant_mass():
    H = Hamiltonian(10, 1, 1, spatial_scheme="nine-point", verbose=False)
    plt.figure()
    plt.title("Constant mass hamiltonian")
    plt.imshow(H.asarray())
    plt.show()

def test_constant_mass_2D():
    H = Hamiltonian((10,30), (1,20), 1e-65, spatial_scheme="seven-point", verbose=False)
    plt.figure()
    plt.title("Constant mass hamiltonian")
    H = H.asarray()
    H[H != 0] = np.log(abs(H[H != 0]))
    plt.imshow(H)
    plt.show()

def test_array_mass():
    N = 200

    # Source: 
    # https://www.researchgate.net/publication/260544509_Band-gap_shift_in_heavily_doped_n-type_Al03Ga07As_alloys
    m_algaas    = 0.0916
    m_gaas      = 0.0665
    V_algaas    = 1.947
    V_gaas      = 1.5194

    n = 3 # odd number
    kernel = np.ones((n)) / n

    m = m_algaas * np.ones(N)
    m[N // 2 - N // 6 : N // 2 + N // 6] = m_gaas
    m *= const.m_e
    m = convolve(m, kernel, mode="same")

    V = V_algaas * np.ones(N)
    V[N // 2 - N // 6 : N // 2 + N // 6] = V_gaas
    V *= const.e_0
    V = convolve(V, kernel, mode="same")

    H = Hamiltonian(N, 20e-9, m, verbose=False)
    H.V = V

    H.plot_potential()

    plt.figure()
    plt.suptitle("Al$_{0.3}$Ga$_{0.7}$As / GaAs / Al$_{0.3}$Ga$_{0.7}$As")
    plt.subplot(1,2,1)
    plt.imshow(H.asarray())
    plt.subplot(1,2,2)
    plt.plot(m)
    plt.ylabel("Mass [kg]")
    plt.show()

    H.plot_eigen(4)

    _, psi = H.eigen(4)
    plt.figure()
    plt.plot(psi[1] - np.roll(psi[1], 1))
    plt.show()

def test_potential():
    N = 100
    L = 2e-9

    r = np.ones(N)
    m = const.m_e * r
    n = 5

    H0 = Hamiltonian(N, L, m, spatial_scheme="three-point", verbose=False)
    z = np.linspace(-L/2, L/2, N)
    V = 100*z**2
    H0.V = V

    E0, psi0 = H0.eigen(n)

    plt.figure()
    plt.title("psi^2")
    for i in range(n):
        plt.subplot(2, 3, i+1)
        plt.title(f"E{i} = {E0[i] / const.e_0 :.3f} eV")
        plt.yticks([])
        plt.plot(psi0[i].real)
    plt.subplot(2, 3, 6)
    plt.plot(V)
    plt.title("Potential")
    plt.tight_layout()
    plt.show()

def test_eigen():
    N = 100
    L = 2e-9

    r = np.ones(N)
    m = const.m_e * r
    n = 4

    H0 = Hamiltonian(N, L, m, spatial_scheme="three-point", verbose=False)
    E0, psi0 = H0.eigen(n)

    H1 = Hamiltonian(N, L, m, spatial_scheme="nine-point", verbose=False)
    E1, psi1 = H1.eigen(n)

    plt.figure()
    plt.title("psi^2")
    for i in range(n):
        plt.subplot(2, 2, i+1)
        plt.title(f"dE{i} = {E1[i] / E0[i] :.3e}")
        plt.plot(abs(psi0[i])**2)
        plt.plot(abs(psi1[i])**2)
    plt.show()

def test_temporal():
    N = 200
    L = 2e-9
    m = const.m_e
    t_end = 10e-15
    dt = 1e-17

    H: list[Hamiltonian] = []
    schemes = ["leapfrog", "crank-nicolson", "scipy-Runge-Kutta 3(2)", "scipy-DOP853"]
    for scheme in schemes:
        H.append(Hamiltonian(N, L, m, temporal_scheme=scheme, verbose=False))

    z = np.linspace(-L/2, L/2, N)
    Vt = lambda t: 6*z**2 + 3*z*np.abs(z)*np.sin(4e15*t)
    for Hi in H:
        Hi.V = Vt

    _, _psi_0 = H[0].eigen(2)
    psi_0 = 2**0-5 * (_psi_0[0, :] + _psi_0[1, :])
    t = []
    psi = []
    for Hi in H:
        ti, psii = Hi.temporal_evolution(0, t_end, dt, psi_0)
        t.append(ti)
        psi.append((abs(psii)**2).real)

    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)

    V = np.array([Vt(tn) for tn in t[0]])

    psi_plots = []
    for psii in psi:
        ln_psi, = ax1.plot(z / 1e-9, psii[0, :])
        psi_plots.append(ln_psi)
    ln_V, = ax2.plot(z / 1e-9, V[0, :] / const.e_0)

    def init():
        ax1.set_title(f"$|\Psi|^2$")
        ax1.set_xlabel("z [nm]")
        ax1.legend(schemes)
        ax2.set_title("Potential [eV]")
        ax2.set_xlabel("z [nm]")
        ax1.set_ylim(0, np.max(psi[0]) * 1.1)
        ax2.set_ylim(np.min(V / const.e_0), np.max(V / const.e_0))
        fig.tight_layout()
        return *psi_plots, ln_V,

    def frames():
        for n in range(psi[0].shape[0]):
            yield *[psii[n, :] for psii in psi], V[n, :]

    def update(data):
        for i, psi_plot in enumerate(psi_plots):
            psi_plot.set_data(z / 1e-9, data[i])
        ln_V.set_data(z / 1e-9, data[-1] / const.e_0)
        return *psi_plots, ln_V,
    
    ani = FuncAnimation(fig, update, frames=frames, init_func=init, blit=False, interval=50)
    plt.show()



"""

Use the scipy backend eigen solver to skip some overhead

"""
import numpy as np
from scipy import sparse as sp
from scipy._lib._threadsafety import ReentrancyLock
from scipy.sparse.linalg import eigsh
from scipy.sparse.linalg._eigen.arpack.arpack import _SymmetricArpackParams


def scipy_get_eigen(
    mat: sp.dia_matrix, n: int, shape: tuple[int], **kwargs
) -> tuple[np.ndarray, np.ndarray]:
    """Calculate :code:`n` eigenvalues of :code:`mat`. Reshape output to :code:`shape`

    :param mat: Matrix to calculate eigenvectors and -values for
    :type mat: sp.dia_matrix
    :param n: Amount of eigenvectors and -values to calculate
    :type n: int
    :param shape: Output shape for eigenvectors
    :type shape: tuple[int]
    :return: eigenvalues, eigenvectors
    :rtype: tuple[np.ndarray(shape = (:code:`n`)), np.ndarray(shape = (:code:`n`, :code:`shape`)]
    """
    if kwargs.get("sigma") is None:
        v, w = _eigsh(mat._mul_vector, mat.shape[0], mat.dtype, k=n, **kwargs)
    else:
        # Fallback to default solver
        v, w = eigsh(mat, k=n, which="SA", **kwargs)

    # Reshape into system shape.
    # Arrays returned from eigsh are fortran ordered
    w = np.array([w[:, i].reshape(shape, order="F") for i in range(n)])
    return v, w


def _eigsh(
    A_OP,
    n,
    dtype,
    k=6,
    sigma=None,
    which="SA",
    v0=None,
    ncv=None,
    maxiter=None,
    tol=0,
    return_eigenvectors=True,
):
    """Copied from the scipy sourcecode, removing some overhead.
    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.eigsh.html
    """
    mode = 1
    M_matvec = None
    Minv_matvec = None
    params = _SymmetricArpackParams(
        n,
        k,
        dtype.char,
        A_OP,
        mode,
        M_matvec,
        Minv_matvec,
        sigma,
        ncv,
        v0,
        maxiter,
        which,
        tol,
    )

    with ReentrancyLock(
        "Nested calls to eigs/eighs not allowed: ARPACK is not re-entrant"
    ):
        while not params.converged:
            params.iterate()

        return params.extract(return_eigenvectors)



import numpy as np
from scipy import sparse as sp


def nabla(
    N: tuple[int],
    L: tuple[float],
    order: int = 2,
    dtype: type = np.float64,
    boundary_condition: str = "zero",
) -> sp.dia_matrix:
    """Finite difference derivative in cartesian coordinates.
    Uses central stencil.

    Example: approximate the derivative of sin(x).

    >>> from qm_sim.spatial_derivative.cartesian import nabla
    >>> import numpy as np
    >>> N = (1000,)
    >>> L = (2*np.pi,)
    >>> n = nabla( N, L, boundary_condition="periodic")
    >>> x = np.linspace( 0, L[0], N[0], endpoint=False )
    >>> y = np.sin(x)
    >>> np.allclose(n @ y, np.cos(x)) # The analytical solution is cos(x)
    True

    :param N: Discretization count along each axis
    :type N: tuple[int]
    :param L: System size along each axis
    :type L: tuple[float]
    :param order: Numerical order of the differentiation scheme.
        Options are:

            - 2
            - 4
            - 6
            - 8

        Defaults to 2.
    :type order: int, optional
    :param dtype: datatype of matrix.
        Defaults to np.float64
    :type dtype: type, optional
    :raises NotImplementedError: if requested order is not available
    :return: Discretized derivative matrix
    :rtype: sp.dia_matrix
    :param boundary_condition: Which boundary condition to apply.
            Options are:

            - zero
            - periodic

            Defaults to "zero"
    :type boundary_condition: str, optional
    """
    # Lookup for first half of stencil.
    # A 0 is appended for the central element,
    # and the rest is mirrored with a sign change later
    match order:
        case 2:
            stencil = [-1 / 2]
        case 4:
            stencil = [1 / 12, -2 / 3]
        case 6:
            stencil = [-1 / 60, 3 / 20, -3 / 4]
        case 8:
            stencil = [1 / 280, -4 / 105, 1 / 5, -4 / 5]
        case _:
            raise NotImplementedError(
                f"Finite difference scheme not found for {order = }"
            )
    stencil += [0]

    stencil = _mirror_sign_list(stencil)
    return _matrix_from_central_stencil(stencil, 1, N, L, dtype, boundary_condition)


def laplacian(
    N: tuple[int],
    L: tuple[float],
    order: int = 2,
    dtype: type = np.float64,
    boundary_condition: str = "zero",
) -> sp.dia_matrix:
    """Finite difference double derivative in cartesian coordinates.
    Uses central stencil

    Example: approximate the second derivative of sin(x).

    >>> from qm_sim.spatial_derivative.cartesian import laplacian
    >>> import numpy as np
    >>> N = (1000,)
    >>> L = (2*np.pi,)
    >>> n = laplacian( N, L, boundary_condition="periodic" )
    >>> x = np.linspace( 0, L[0], N[0], endpoint=False)
    >>> y = np.sin(x)
    >>> np.allclose(n @ y, -np.sin(x)) # The analytical solution is -sin(x)
    True

    :param N: Discretization count along each axis
    :type N: tuple[int]
    :param L: System size along each axis
    :type L: tuple[float]
    :param order: Numerical order of the differentiation scheme.
        Options are:

            - 2
            - 4
            - 6
            - 8

        Defaults to 2.
    :type order: int, optional
    :param dtype: datatype of matrix.
        Defaults to np.float64
    :type dtype: type, optional
    :raises NotImplementedError: if requested order is not available
    :return: Discretized derivative matrix
    :rtype: sp.dia_matrix
    :param boundary_condition: Which boundary condition to apply.
            Options are:

            - zero
            - periodic

            Defaults to "zero"
    :type boundary_condition: str, optional
    """
    # lookup the first half of the stencil, to be mirrored later
    match order:
        case 2:
            stencil = [1, -2]
        case 4:
            stencil = [-1 / 12, 4 / 3, -5 / 2]
        case 6:
            stencil = [1 / 90, -3 / 20, 3 / 2, -49 / 18]
        case 8:
            stencil = [-1 / 560, 8 / 315, -1 / 5, 8 / 5, -205 / 72]
        case _:
            raise NotImplementedError(
                f"Finite difference scheme not found for {order = }"
            )

    stencil = _mirror_list(stencil)
    return _matrix_from_central_stencil(stencil, 2, N, L, dtype, boundary_condition)


def _matrix_from_central_stencil(
    stencil: list[float],
    power: int,
    N: tuple[int],
    L: tuple[float],
    dtype: type,
    boundary_condition: str = "zero",
) -> sp.dia_matrix:
    """
    Creates a full matrix from a central stencil. Determines indices from stencil
    """

    available_boundary_conditions = ["zero", "periodic"]
    if boundary_condition not in available_boundary_conditions:
        raise ValueError(
            f"Invalid boundary condition: {boundary_condition}. Options are: "
            + ", ".join(available_boundary_conditions)
        )

    if boundary_condition == "zero":
        indices = np.arange(len(stencil)) - len(stencil) // 2
        axis_indices = [indices for _ in N]

    elif boundary_condition == "periodic":
        indices = np.arange(len(stencil)) - len(stencil) // 2
        # There might be different sizes for each axis.
        # Therefore, we keep the updated indices for each axis
        periodic_indices = [list() for _ in N]
        periodic_stencil = []
        for i, ind in enumerate(indices):
            # Lower diagonals
            if ind < 0:
                for i, n in enumerate(N):
                    periodic_indices[i].append(-n - ind)
                periodic_stencil.append(stencil[i])
            # Upper diagonals
            elif ind > 0:
                for i, n in enumerate(N):
                    periodic_indices[i].append(n - ind)
                periodic_stencil.append(stencil[i])
            # Also keep the central diagonals
            else:
                for i in range(len(N)):
                    periodic_indices[i] += list(indices)
                periodic_stencil += stencil
        axis_indices = np.array(periodic_indices)
        stencil = np.array(periodic_stencil)

    mat = np.zeros((1, 1))
    for l, n, indices in zip(L, N, axis_indices):
        h = l / n
        next_mat = sp.diags(
            stencil,
            indices,
            shape=(n, n),
            dtype=dtype,
            format="dia",
        )
        next_mat *= 1 / h**power
        mat = sp.kronsum(mat, next_mat, format="dia")

    return mat


def _mirror_list(l: list) -> list:
    """
    Mirror a list around the final element.

    Example: [a, b, c] -> [a, b, c, b, a]
    """
    return l + l[-2::-1]


def _mirror_sign_list(l: list) -> list:
    """
    Mirror a list around the final element,
    and flip the sign of each new element.

    Example: [-a, b, c] -> [-a, b, c, -b, a]
    """
    return l + [-i for i in l[-2::-1]]



"""

Script to generate an __init__.py-file for the C++ module

"""

import pathlib
import sys


def main(module_name: str):
    path = pathlib.Path(__file__).parent.parent
    with open(path / "__init__.py", "w") as file:
        file.write(
            f"""try:
    from .{module_name} import *
except ImportError:
    print("Warning: {module_name} C++ module not found")
"""
        )


if __name__ == "__main__":
    if len(sys.argv) < 2:
        raise Exception("No module name supplied. Please enter a module name.")
    main(sys.argv[1])



"""

Plots for the solutions to the Schrödinger equation

"""
from typing import Callable

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import ArtistAnimation

from .nature_constants import e_0


def _get_plot_fun(ndim: int, ax: plt.Axes = None):
    if ax is None:
        ax = plt

    if ndim == 1:
        return lambda *args, **kwargs: ax.plot(*args, c="b", **kwargs)
    elif ndim == 2:
        # To preserve the return type of :code:`plot`above
        return lambda *args, **kwargs: [ax.imshow(*args, **kwargs)]
    elif ndim == 3:
        raise NotImplementedError("3D plots not yet supported")
    else:
        # It would be impressive if this is ever executed
        raise ValueError(f"Invalid system dimensionality: {ndim}D")


def _shape_from_int(n: int) -> tuple[int, int]:
    """Get plot shape from n (plot count)"""
    if n < 1:
        raise ValueError("Plot count must be positive")
    shapes = {
        1: (1, 1),
        2: (1, 2),
        3: (1, 3),
        4: (2, 2),
        5: (2, 3),
        6: (2, 3),
        7: (2, 4),
        8: (2, 4),
        9: (3, 3),
        10: (3, 4),
        11: (3, 4),
        12: (3, 4),
    }
    return shapes.get(n, (1, n))


def eigen(E: np.ndarray, psi: np.ndarray):
    """Plot absolute square of wave function

    :param E: Eigenenergies
    :type E: np.ndarray
    :param psi: Eigenfunctions
    :type psi: np.ndarray
    """
    n = E.shape[0]
    ndim = len(psi.shape) - 1
    shape = _shape_from_int(n)

    plt.figure()
    plt.suptitle("$|\\Psi|^2$")
    plot = _get_plot_fun(ndim)
    for i in range(n):
        plt.subplot(*shape, i + 1)
        plt.title(f"E{i} = {E[i] / e_0 :.3f} eV")
        plot(abs(psi[i]) ** 2)
    plt.tight_layout()
    plt.show()
    return


def temporal(t: np.ndarray, psi: np.ndarray, Vt: Callable[[float], np.ndarray]):
    """Create an animation of the wave function, alongside the potential

    :param t: Times correspondind to the wave functions
    :type t: np.ndarray
    :param psi: Wave funcions
    :type psi: np.ndarray
    :param Vt: Function of time, returning the potential at that time
    :type Vt: Callable[[float], np.ndarray]
    """
    ndim = len(psi.shape) - 1

    # Get potential at each timestep
    V = np.array([Vt(tn) for tn in t])

    # We want to plot the probability distribution
    psi2 = np.abs(psi) ** 2

    # Plot the results
    fig, (ax1, ax2) = plt.subplots(2, 1)
    ax1_plot = _get_plot_fun(ndim, ax1)
    ax2_plot = _get_plot_fun(ndim, ax2)

    (psi_plot,) = ax1_plot(psi2[0, :])
    (V_plot,) = ax2_plot(V[0, :] / e_0)

    ax1.set_title(f"$|\\Psi|^2$")
    ax2.set_title("Potential [eV]")
    if ndim == 1:
        ax1.set_ylim(0, np.max(psi2) * 1.1)
        ax2.set_ylim(np.min(V / e_0), np.max(V / e_0))
    elif ndim == 2:
        ax1.set_xticks([])
        ax1.set_yticks([])
        ax2.set_xticks([])
        ax2.set_yticks([])
    fig.tight_layout()

    ims = [(psi_plot, V_plot)]
    for n in range(psi2.shape[0]):
        (psi_plot,) = ax1_plot(psi2[n, ...], animated=True)
        (V_plot,) = ax2_plot(V[n, ...] / e_0, animated=True)
        ims.append(
            (
                psi_plot,
                V_plot,
            )
        )

    ani = ArtistAnimation(fig, ims, blit=True, interval=50)
    plt.show()
    return


def potential(V: np.ndarray):
    """Plot the potential

    :param V: Potential array
    :type V: np.ndarray
    """
    plt.figure()
    _get_plot_fun(len(V.shape))(V / e_0)
    plt.title("Potential [eV]")
    plt.show()
    return
